import arcade, math, copy, pymunk, warnings

from .collision import collide

from .point import Rotate_Points
from .point import Scale_X_Points
from .point import Scale_Y_Points
from .point import Center_X_Of_Points
from .point import Center_Y_Of_Points
from .point import Left_Of_Points
from .point import Top_Of_Points
from .point import Right_Of_Points
from .point import Bottom_Of_Points

from .math import Angle
from .math import Distance
from .math import Direction

class Pymunk:
    
    def __init__(self):
        self.engine = None
        self.shape = None
        self.body = None
        self.moment = None
        self.max_x_speed = None
        self.max_y_speed = None
        self.x_gravity = None
        self.y_gravity = None
        self.damping = None
        self.joint = None

class Base(object):

    def collide(self, other):
        if self.pymunk.engine is other.pymunk.engine is not None:
            return self.pymunk.engine.collide(self, other)
            
        return collide(self.boundary(), other.boundary())
    
    def copy(self):
        return copy.deepcopy(self)
    
    def angle(self, other):
        return Angle(self.x, self.y, other.x, other.y)

    def distance(self, other):
        return Distance(self.x, self.y, other.x, other.y)
    
    def on_ground(self):# This is a copied, complicated Arcade function.
        grounding = {'normal': pymunk.Vec2d.zero(),
                     'penetration': pymunk.Vec2d.zero(),
                     'impulse': pymunk.Vec2d.zero(),
                     'position': pymunk.Vec2d.zero(),
                     'body': None}

        def f(arbiter):
            n = -arbiter.contact_point_set.normal
            
            if n.y > grounding['normal'].y:
                
                grounding['normal'] = n
                grounding['penetration'] = -arbiter.contact_point_set.points[0].distance
                grounding['body'] = arbiter.shapes[1].body
                grounding['impulse'] = arbiter.total_impulse
                grounding['position'] = arbiter.contact_point_set.points[0].point_b

        self.pymunk.body.each_arbiter(f)

        return grounding['body'] is not None
    
    def apply_impulse(self, x_impulse: float, y_impulse: float,
                      x_point: float = 0, y_point: float = 0):
        self.pymunk.body.apply_impulse_at_local_point((x_impulse, y_impulse),
                                                      (x_point, y_point))
        
    def apply_force(self, x_force: float, y_force: float, x_point: float = 0,
                    y_point: float = 0):
        
        self.pymunk.body.apply_force_at_local_point((x_force, y_force),
                                                    (x_point, y_point))    
        
    def _get_max_x_speed(self):
        return self.pymunk.max_x_speed
    
    def _set_max_x_speed(self, value: float):
        self.pymunk.max_x_speed = value
        
    max_x_speed = property(_get_max_x_speed, _set_max_x_speed)
    
    def _get_max_y_speed(self):
        return self.pymunk.max_y_speed
    
    def _set_max_y_speed(self, value: float):
        self.pymunk.max_y_speed = value
        
    max_y_speed = property(_get_max_y_speed, _set_max_y_speed)
    
    def _get_x_speed(self):
        return self.pymunk.body.velocity[0]
    
    def _set_x_speed(self, value: float):
        self.pymunk.body.velocity = (value, self.pymunk.body.velocity[1])
        
    x_speed = property(_get_x_speed, _set_x_speed)
    
    def _get_y_speed(self):
        return self.pymunk.body.velocity[1]
    
    def _set_y_speed(self, value: float):
        self.pymunk.body.velocity = (self.pymunk.body.velocity[0], value)
        
    y_speed = property(_get_y_speed, _set_y_speed)
        
    def kill(self):
        if self.array is not None:
            self.array.remove(self)
            
        if self.pymunk.engine is not None:
            self.pymunk.engine.remove(self)
            
        del self
        
    def _get_x(self):
        return self.sprite.center_x
    
    def _set_x(self, value: float):
        if self.pymunk.engine is not None:
            self.pymunk.body.position = value, self.pymunk.shape.body.position.y
            self.pymunk.engine.space.reindex_shapes_for_body(self.pymunk.shape.body)
            self.x_speed = 0
            
        self.sprite.center_x = value
            
    x = property(_get_x, _set_x)

    def _get_y(self):
        return self.sprite.center_y
    
    def _set_y(self, value: float):
        if self.pymunk.engine is not None:
            self.pymunk.body.position = self.pymunk.shape.body.position.x, value
            self.pymunk.engine.space.reindex_shapes_for_body(self.pymunk.body)
            self.y_speed = 0
            
        self.sprite.center_y = value
            
    y = property(_get_y, _set_y)
    
    def _get_rotation(self):
        return self.sprite.angle
    
    def _set_rotation(self, value: float):
        if self.pymunk.engine is not None:
            self.pymunk.body.angle = math.radians(value)
            self.pymunk.engine.space.reindex_shapes_for_body(self.pymunk.body)
            
        self.sprite.angle = value
    
    rotation = property(_get_rotation, _set_rotation)
    
    def _get_x_gravity(self):
        return self.pymunk.x_gravity
    
    def _set_x_gravity(self, value: float):
        self.pymunk.x_gravity = value
        
    x_gravity = property(_get_x_gravity, _set_x_gravity)
    
    def _get_y_gravity(self):
        return self.pymunk.y_gravity
    
    def _set_y_gravity(self, value: float):
        self.pymunk.y_gravity = value
        
    y_gravity = property(_get_y_gravity, _set_y_gravity)
    
    def _get_damping(self):
        return self.pymunk.damping
    
    def _set_damping(self, value: float):
        self.pymunk.damping = value
        
    damping = property(_get_damping, _set_damping)
    
    def _get_mass(self):
        return self.pymunk.body.mass
    
    def _set_mass(self, value: float):
        self.pymunk.body.mass = value
        self.pymunk.shape.mass = value
        
    mass = property(_get_mass, _set_mass)
    
    def _get_body(self):
        return self.pymunk.body.body_type
    
    def _set_body(self, value: float):
        self.pymunk.body.body_type = value
        
    body = property(_get_body, _set_body)
    
    def _get_elasticity(self):
        return self.pymunk.shape.elasticity
    
    def _set_elasticity(self, value: float):
        self.pymunk.shape.elasticity = value
        
    elasticity = property(_get_elasticity, _set_elasticity)
    
    def _get_moment(self):
        return self.pymunk.body.moment
    
    def _set_moment(self, value: float):
        self.pymunk.body.moment = value
        self.pymunk.shape.moment = value
        
    moment = property(_get_moment, _set_moment)
    
    def _get_friction(self):
        return self.pymunk.shape.friction
    
    def _set_friction(self, value: float):
        self.pymunk.shape.friction = value
        
    friction = property(_get_friction, _set_friction)    
    
class Point(Base):
    
    def __init__(self, x: float = 100, y: float = 100, color: arcade.Color = (0, 0, 0),
                 size: float = 10, rotation: float = 0):

        self._color = color
        self._size = size
        self.pymunk = Pymunk()
        self.array = None
        
        self.update(x, y, rotation)
        
    def draw(self):
        self.sprite.program['Angle'] = self.sprite._angle
        self.sprite.draw()
        
    def boundary(self):
        return {'points': self.get_corners(),
                'type': 'polygon'}
    
    def get_corners(self):
        return arcade.get_rectangle_points(self.sprite.center_x,
                                           self.sprite.center_y,  self._size,
                                           self._size)
    
    def update(self, x = None, y = None, angle = None):
        
        if x is None: x = self.sprite.center_x
        if y is None: y = self.sprite.center_y
        if angle is None: angle = self.sprite.angle
            
        self.shape = arcade.create_rectangle_filled(0, 0, self._size,
                                                    self._size, self._color)
        
        self.sprite = arcade.ShapeElementList()
        self.sprite.append(self.shape)
        
        self.sprite.center_x = x
        self.sprite.center_y = y
        self.sprite.angle = angle
        
    def get_moment(self, mass):
        return pymunk.moment_for_box(mass, (self._size, self._size))
    
    def get_shape(self, body):
        return pymunk.Poly.create_box(body, (self._size, self._size))
        
    def _get_top(self):
        return Top_Of_Points(self.get_corners())
    
    def _set_top(self, value: float):
        distance = value - self.top
        
        self.sprite.center_y += distance
        
    top = property(_get_top, _set_top)
    
    def _get_bottom(self):
        return Bottom_Of_Points(self.get_corners())
    
    def _set_bottom(self, value: float):
        distance = value - self.bottom
        
        self.sprite.center_y += distance
        
    bottom = property(_get_bottom, _set_bottom)
    
    def _get_left(self):
        return Left_Of_Points(self.get_corners())
    
    def _set_left(self, value: float):
        distance = value - self.left
        
        self.sprite.center_x += distance
        
    left = property(_get_left, _set_left)
    
    def _get_right(self):
        return Right_Of_Points(self.get_corners())
    
    def _set_right(self, value: float):
        distance = value - self.right
        
        self.sprite.center_x += distance
        
    right = property(_get_right, _set_right)
    
    def _get_size(self):
        return self._size
    
    def _set_size(self, value: float):
        if self._size is not value:
            self._size = value
        
            self.update()
            
            if self.pymunk.engine is not None:
                self.pymunk.engine.refresh(self)            
            
    size = property(_get_size, _set_size)
    
    def _get_color(self):
        return self._color
    
    def _set_color(self, value: arcade.Color):
        if self._color is not value:
            self._color = value
        
            self.update()
            
    color = property(_get_color, _set_color)
       
class Line(Base):
    
    def __init__(self, x1: float = 0, y1: float = 0, x2: float = 100,
                 y2: float = 100, thickness: float = 1,
                 color: arcade.Color = (0, 0, 0), rotation: float = 0):
        
        x = (x2 + x1) / 2
        y = (y2 + y1) / 2

        self._x1 = x1
        self._y1 = y1
        self._x2 = x2
        self._y2 = y2
        self._thickness = thickness
        self._color1 = color
        self._color2 = color
        self.pymunk = Pymunk()
        self.array = None

        self.update(x, y, rotation)
        
    def draw(self):
        self.sprite.program['Angle'] = self.sprite._angle
        self.sprite.draw()
        
    def boundary(self):
        return {'points': self.get_corners(),
                'type': 'polygon'}
    
    def get_corners(self):
        points = self.get_current()
        points = Rotate_Points(((points[0][0] + self.sprite.center_x,
                                 points[0][1] + self.sprite.center_y),
                                (points[1][0] + self.sprite.center_x,
                                 points[1][1] + self.sprite.center_y)),
                                self.sprite.angle, self.sprite.center_x,
                                self.sprite.center_y)

        return arcade.get_points_for_thick_line(points[0][0], points[0][1],
                                                points[1][0], points[1][1],
                                                self._thickness)
    def get_current(self):
        x = (self._x2 + self._x1) / 2
        y = (self._y2 + self._y1) / 2
                        
        return (self._x1 - x, self._y1 - y), (self._x2 - x, self._y2 - y)
    
    def update(self, x = None, y = None, angle = None):
                
        if x is None: x = self.sprite.center_x
        if y is None: y = self.sprite.center_y
        if angle is None: angle = self.sprite.angle
                
        self.shape = arcade.create_lines_with_colors(
            self.get_current(), [self._color1, self._color2],
            line_width = self._thickness)
        
        self.sprite = arcade.ShapeElementList()
        self.sprite.append(self.shape)
        
        self.sprite.center_x = x
        self.sprite.center_y = y
        self.sprite.angle = angle
            
    def get_moment(self, mass):
        points = self.get_current()
        
        return pymunk.moment_for_poly(mass, arcade.get_points_for_thick_line(
            points[0][0], points[0][1], points[1][0], points[1][1],
            self._thickness))
    
    def get_shape(self, body):
        points = self.get_current()
        
        return pymunk.Poly(body, arcade.get_points_for_thick_line(
            points[0][0], points[0][1], points[1][0], points[1][1],
            self._thickness))
    
    def _get_points(self):
        return ((self._x1, self._y1), (self._x2, self._y2))
    
    def _set_points(self, value: arcade.PointList):        
        self._x1 = value[0][0]
        self._y1 = value[0][1]
        self._x2 = value[1][0]
        self._y2 = value[1][1]
                        
        self.update((self._x2 + self._x1) / 2, (self._y2 + self._y1) / 2)
        
        if self.pymunk.engine is not None:
            self.pymunk.engine.refresh(self)        
        
    points = property(_get_points, _set_points)
    
    def _get_x1(self):
        return self._x1
    
    def _set_x1(self, value: float):
        if self._x1 is not value:
            self._x1 = value
                
            self.update(x = (self._x2 + self._x1) / 2)
            
            if self.pymunk.engine is not None:
                self.pymunk.engine.refresh(self)            
        
    x1 = property(_get_x1, _set_x1)
    
    def _get_y1(self):
        return self._y1
    
    def _set_y1(self, value: float):
        if self._y1 is not value:
            self._y1 = value
        
            self.update(y = (self._y2 + self._y1) / 2)
            
            if self.pymunk.engine is not None:
                self.pymunk.engine.refresh(self)            
                
    y1 = property(_get_y1, _set_y1)
    
    def _get_x2(self):
        return self._x2
    
    def _set_x2(self, value: float):
        if self._x2 is not value:
            self._x2 = value
        
            self.update(x = (self._x2 + self._x1) / 2)
            
            if self.pymunk.engine is not None:
                self.pymunk.engine.refresh(self)            
        
    x2 = property(_get_x2, _set_x2)
    
    def _get_y2(self):
        return self._y2
    
    def _set_y2(self, value: float):
        if self._y2 is not value:
            self._y2 = value
        
            self.update(y = (self._y2 + self._y1) / 2)
            
            if self.pymunk.engine is not None:
                self.pymunk.engine.refresh(self)            
        
    y2 = property(_get_y2, _set_y2)
    
    def _get_color(self):
        if self._color1 == self._color2:
            return self._color1
        
        return self._color1, self._color2
    
    def _set_color(self, value: arcade.Color):
        if self._color1 is not value or self._color2 is not value:
            self._color1 = value
            self._color2 = value
        
            self.update()
        
    color = property(_get_color, _set_color)
    
    def _get_color1(self):
        return self._color1
    
    def _set_color1(self, value: arcade.Color):
        if self._color1 is not value:
            self._color1 = value
        
            self.update()
        
    color1 = property(_get_color1, _set_color1)
    
    def _get_color2(self):
        return self._color2
    
    def _set_color2(self, value: arcade.Color):
        if self._color2 is not value:
            self._color2 = value
        
            self.update()
        
    color2 = property(_get_color2, _set_color2)    
    
    def _get_thickness(self):
        return self._thickness
    
    def _set_thickness(self, value: float):
        if self._thickness is not value:
            self._thickness = value
        
            self.update()
        
            if self.pymunk.engine is not None:
                self.pymunk.engine.refresh(self)
        
    thickness = property(_get_thickness, _set_thickness)

    def _get_top(self):
        return Top_Of_Points(self.get_corners())
    
    def _set_top(self, value: float):
        distance = value - self.top
        
        self.sprite.center_y += distance
            
    top = property(_get_top, _set_top)
    
    def _get_bottom(self):
        return Bottom_Of_Points(self.get_corners())
    
    def _set_bottom(self, value: float):
        distance = value - self.bottom
        
        self.sprite.center_y += distance
            
    bottom = property(_get_bottom, _set_bottom)
    
    def _get_left(self):
        return Left_Of_Points(self.get_corners())
    
    def _set_left(self, value: float):
        distance = value - self.left
        
        self.sprite.center_x += distance
            
    left = property(_get_left, _set_left)
    
    def _get_right(self):
        return Right_Of_Points(self.get_corners())
    
    def _set_right(self, value: float):
        distance = value - self.right
        
        self.sprite.center_x += distance
            
    right = property(_get_right, _set_right)
    
class Position:
    
    def __init__(self, point, shape):
        self.array = list(point)
        self.shape = shape
        
    def __getitem__(self, key):
        return self.array[key]    
        
    def __setitem__(self, key, value):
        self.array[key] = value
        
        self.shape.update(Center_X_Of_Points(self.shape.points),
                          Center_Y_Of_Points(self.shape.points))
        
        if self.shape.pymunk.engine is not None:
            self.pymunk.engine.refresh(self)
            
    def __iter__(self):
        return iter(self.array)
    
    def __str__(self):
        return str(self.array)
    
    def __repr__(self):
        return repr(self.array)
    
    def __len__(self):
        return len(self.array)
    
class Positions(Position):
    
    def __init__(self, points, shape):
        self.array = [Position(point, shape) for point in points]
        self.shape = shape
    
class Strip(Base):
    
    def __init__(self,
                 points: arcade.PointList = ((0, 0), (100, 100), (150, 50)),
                 color: arcade.Color = (0, 0, 0), thickness: float = 1,
                 rotation: float = 0):

        self.points = Positions(points, self)
        self._color = color
        self._thickness = thickness
        self.pymunk = Pymunk()
        self.array = None

        self.update(Center_X_Of_Points(self._points),
                    Center_Y_Of_Points(self._points), rotation)
        
    def draw(self):
        self.sprite.program['Angle'] = self.sprite._angle
        self.sprite.draw()
        
    def boundary(self):
        return {'points': self.get_corners(),
                'type': 'polygon'}
    
    def get_moment(self, mass):
        return pymunk.moment_for_poly(mass, self.get_current())
    
    def get_shape(self, body):
        return pymunk.Poly(body, self.get_current())
    
    def get_corners(self):
        """
        I don't know how to get all corners of the strip. The strip
        collision isn't accurate. Is there an Arcade function similar to
        get_points_from_thick_line that applies to line strips?
        """
        positions = []
        
        for point in self.get_current():
            position = (point[0] + self.sprite.center_x,
                        point[1] + self.sprite.center_y)
                    
            positions.append(position)
        
        return Rotate_Points(positions, self.sprite.angle,
                             self.sprite.center_x, self.sprite.center_y)
    
    def get_current(self):
        x = Center_X_Of_Points(self._points)
        y = Center_Y_Of_Points(self._points)        
        
        positions = []
        
        for point in self._points:
            positions.append([point[0] - x, point[1] - y])
            
        return positions
    
    def update(self, x = None, y = None, angle = None):
        
        if x is None: x = self.sprite.center_x
        if y is None: y = self.sprite.center_y
        if angle is None: angle = self.sprite.angle
            
        self.shape = arcade.create_line_strip(self.get_current(), self._color,
                                              self._thickness)
        
        self.sprite = arcade.ShapeElementList()
        self.sprite.append(self.shape)
        
        self.sprite.center_x = x
        self.sprite.center_y = y
        self.sprite.angle = angle    
    
    def _get_points(self):
        return self._points
    
    def _set_points(self, value: arcade.PointList):
        self._points = Positions(value, self)
        
        self.update(Center_X_Of_Points(self._points),
                    Center_Y_Of_Points(self._points))
        
        if self.pymunk.engine is not None:
            self.pymunk.engine.refresh(self)        
    
    points = property(_get_points, _set_points)    
    
    def _get_top(self):
        return Top_Of_Points(self.get_corners())
    
    def _set_top(self, value: float):
        distance = value - self.top
        
        self.sprite.center_y += distance
            
    top = property(_get_top, _set_top)
    
    def _get_bottom(self):
        return Bottom_Of_Points(self.get_corners())
    
    def _set_bottom(self, value: float):
        distance = value - self.bottom
        
        self.sprite.center_y += distance
            
    bottom = property(_get_bottom, _set_bottom)
    
    def _get_left(self):
        return Left_Of_Points(self.get_corners())
    
    def _set_left(self, value: float):
        distance = value - self.left
        
        self.sprite.center_x += distance
            
    left = property(_get_left, _set_left)
    
    def _get_right(self):
        return Right_Of_Points(self.get_corners())
    
    def _set_right(self, value: float):
        distance = value - self.right
        
        self.sprite.center_x += distance
            
    right = property(_get_right, _set_right)
    
    def _get_color(self):
        return self._color
    
    def _set_color(self, value: arcade.Color):
        if self._color is not value:
            self._color = value
            
            self.update()
            
    color = property(_get_color, _set_color)
    
    def _get_thickness(self):
        return self._thickness
    
    def _set_thickness(self, value: bool):
        if self._thickness is not value:
            self._thickness = value
            
            self.update()       
            
    thickness = property(_get_thickness, _set_thickness)
        
class Shape(Base):
    
    def __init__(self,
                 points: arcade.PointList = ((0, 0), (100, 200), (150, 100), (90, 10)),
                 color: arcade.Color = (0, 0, 0), outline: float = 0,
                 rotation: float = 0):

        self._points = Positions(points, self)
        self._color = color
        self._outline = outline
        self.pymunk = Pymunk()
        self.array = None
        
        self.update(Center_X_Of_Points(self._points),
                    Center_Y_Of_Points(self._points), rotation)
        
    def draw(self):
        self.sprite.program['Angle'] = self.sprite._angle
        self.sprite.draw()
        
    def boundary(self):
        return {'points': self.get_corners(),
                'type': 'polygon'}
    
    def get_corners(self):
        positions = []
        
        for point in self.get_current():
            position = (point[0] + self.sprite.center_x,
                        point[1] + self.sprite.center_y)
                    
            positions.append(position)        
        
        return Rotate_Points(positions, self.sprite.angle,
                             self.sprite.center_x, self.sprite.center_y)
    
    def get_current(self):        
        positions = []
        
        x = Center_X_Of_Points(self._points)
        y = Center_Y_Of_Points(self._points)        
        
        for point in self._points:
            positions.append([point[0] - x, point[1] - y])
            
        return positions
    
    def update(self, x = None, y = None, angle = None):
        
        if x is None: x = self.sprite.center_x
        if y is None: y = self.sprite.center_y
        if angle is None: angle = self.sprite.angle
            
        if self.outline:
            self.shape = arcade.create_line_loop(self.get_current(),
                                                 self._color, self._outline)
        else:
            self.shape = arcade.create_polygon(self.get_current(),
                                               self._color)
        
        self.sprite = arcade.ShapeElementList()
        self.sprite.append(self.shape)
        
        self.sprite.center_x = x
        self.sprite.center_y = y
        self.sprite.angle = angle   
        
    def get_moment(self, mass):
        return pymunk.moment_for_poly(mass, self.get_current())
    
    def get_shape(self, body):
        return pymunk.Poly(body, self.get_current())
    
    def _get_points(self):
        return self._points
    
    def _set_points(self, value: arcade.PointList):
        self._points = Positions(value, self)
        
        self.update(Center_X_Of_Points(self._points),
                    Center_Y_Of_Points(self._points))
        
        if self.pymunk.engine is not None:
            self.pymunk.engine.refresh(self)        
        
    points = property(_get_points, _set_points)
    
    def _get_top(self):
        return Top_Of_Points(self.get_corners())
    
    def _set_top(self, value: float):
        distance = value - self.top
        
        self.sprite.center_y += distance
            
    top = property(_get_top, _set_top)
    
    def _get_bottom(self):
        return Bottom_Of_Points(self.get_corners())
    
    def _set_bottom(self, value: float):
        distance = value - self.bottom
        
        self.sprite.center_y += distance
            
    bottom = property(_get_bottom, _set_bottom)
    
    def _get_left(self):
        return Left_Of_Points(self.get_corners())
    
    def _set_left(self, value: float):
        distance = value - self.left
        
        self.sprite.center_x += distance
            
    left = property(_get_left, _set_left)
    
    def _get_right(self):
        return Right_Of_Points(self.get_corners())
    
    def _set_right(self, value: float):
        distance = value - self.right
        
        self.sprite.center_x += distance
            
    right = property(_get_right, _set_right)
    
    def _get_color(self):
        return self._color
    
    def _set_color(self, value: arcade.Color):
        if self._color is not value:
            self._color = value
            
            self.update()
            
    color = property(_get_color, _set_color)
    
    def _get_outline(self):
        return self._outline
    
    def _set_outline(self, value: bool):
        if self._outline is not value:
            self._outline = value
            
            self.update()
            
    outline = property(_get_outline, _set_outline)
        
class Circle(Base):
    
    def __init__(self, x: float = 100, y: float = 100, size: float = 50,
                 color: arcade.Color = (0, 0, 0), outline: float = 0,
                 segments: int = 32):

        self._size = size
        self._color1 = color
        self._color2 = color
        self._outline = outline
        self._segments = segments
        self.pymunk = Pymunk()
        self.array = None

        self.update(x, y, 0)
        
    def draw(self):
        self.sprite.draw()
        
    def boundary(self):
        return {'points': self.get_corners(),
                'type': 'polygon'}
    
    def get_corners(self):
        positions = []
        
        for index in range(1, self._segments):
            x, y = Direction(self._size / 2, index * 360 / self._segments)
            point = [self.sprite.center_x + x, self.sprite.center_y + y]
            
            positions.append(point)
            
        return positions
    
    def get_current(self):
        positions = []
        
        for index in range(1, self._segments):
            x, y = Direction(self._size / 2, index * 360 / self._segments)
            
            positions.append((x, y))
            
        return positions        
        
    def update(self, x = None, y = None, angle = None):
        
        if x is None: x = self.sprite.center_x
        if y is None: y = self.sprite.center_y
        if angle is None: angle = self.sprite.angle
        
        if self._outline:
            self.shape = arcade.create_line_loop(self.get_current(),
                                                 self._color1, self._outline)            
        else:
            self.shape = arcade.create_ellipse_filled_with_colors(
                0, 0, self._size / 2, self._size / 2, self._color1,
                self._color2, num_segments = self._segments)            
        
        self.sprite = arcade.ShapeElementList()
        self.sprite.append(self.shape)
        
        self.sprite.center_x = x
        self.sprite.center_y = y
        self.sprite.angle = angle    
    
    def get_moment(self, mass):
        return pymunk.moment_for_circle(mass, 0, self._size / 2)
    
    def get_shape(self, body):
        return pymunk.Circle(body, self._size / 2)
    
    def _get_top(self):
        return Top_Of_Points(self.get_corners())
    
    def _set_top(self, value: float):
        distance = value - self.top
        
        self.sprite.center_y += distance
            
    top = property(_get_top, _set_top)
    
    def _get_bottom(self):
        return Bottom_Of_Points(self.get_corners())
    
    def _set_bottom(self, value: float):
        distance = value - self.bottom
        
        self.sprite.center_y += distance
            
    bottom = property(_get_bottom, _set_bottom)
    
    def _get_left(self):
        return Left_Of_Points(self.get_corners())
    
    def _set_left(self, value: float):
        distance = value - self.left
        
        self.sprite.center_x += distance
            
    left = property(_get_left, _set_left)
    
    def _get_right(self):
        return Right_Of_Points(self.get_corners())
    
    def _set_right(self, value: float):
        distance = value - self.right
        
        self.sprite.center_x += distance
            
    right = property(_get_right, _set_right)
    
    def _get_size(self):
        return self._size
    
    def _set_size(self, value: float):
        if self._size is not value:
            self._size = value
            
            self.update()
            
            if self.pymunk.engine is not None:
                self.pymunk.engine.refresh(self)            
            
    size = property(_get_size, _set_size)
    
    def _get_color(self):
        if self._color1 == self._color2:
            return self._color1
        
        return self._color1, self._color2
    
    def _set_color(self, value: arcade.Color):
        
        if not self._color1 == self._color2 == value:
            self._color1 = value
            self._color2 = value       
        
            self.update()
            
    color = property(_get_color, _set_color)
    
    def _get_color1(self):
        return self._color1
    
    def _set_color1(self, value: arcade.Color):
        if self._color1 is not value:
            self._color1 = value
        
            self.update()
        
    color1 = property(_get_color1, _set_color1)
    
    def _get_color2(self):
        return self._color2
    
    def _set_color2(self, value: arcade.Color):
        if self._color2 is not value:
            self._color2 = value
        
            self.update()
        
    color2 = property(_get_color2, _set_color2)
    
    def _get_outline(self):
        return self._outline
    
    def _set_outline(self, value: float):
        if self._outline is not value:
            self._outline = value
        
            self.update()
        
    outline = property(_get_outline, _set_outline)
    
    def _get_segments(self):
        return self._segments
    
    def _set_segments(self, value: float):
        if self._segments is not value:
            self._segments = value
        
            self.update()
        
    segments = property(_get_segments, _set_segments)

class Box(Base):
    
    def __init__(self, x: float = 100, y: float = 100, width: float = 50,
                 height: float = 30, color: arcade.Color = (0, 0, 0),
                 outline: float = 0, rotation: float = 0):

        self._width = width
        self._height = height
        self._color1 = color
        self._color2 = color
        self._color3 = color
        self._color4 = color
        self._outline = outline
        self.pymunk = Pymunk()
        self.array = None

        self.update(x, y, rotation)
        
    def draw(self):
        self.sprite.program['Angle'] = self.sprite._angle
        self.sprite.draw()
        
    def boundary(self):
        return {'points': self.get_corners(),
                'type': 'polygon'}
    
    def get_corners(self):
        return arcade.get_rectangle_points(self.sprite.center_x,
                                           self.sprite.center_y, self._width,
                                           self._height, self.sprite.angle)
    
    def update(self, x = None, y = None, angle = None):
        
        if x is None: x = self.sprite.center_x
        if y is None: y = self.sprite.center_y
        if angle is None: angle = self.sprite.angle
        
        if self._outline:
            self.shape = arcade.create_rectangle_outline(0, 0, self._width,
                                                         self._height,
                                                         self._color1,
                                                         self._outline)
        else:
            self.shape = arcade.create_rectangle_filled_with_colors(
                arcade.get_rectangle_points(0, 0, self._width, self._height),
                (self._color1, self._color2, self._color3, self._color4))            
        
        self.sprite = arcade.ShapeElementList()
        self.sprite.append(self.shape)
        
        self.sprite.center_x = x
        self.sprite.center_y = y
        self.sprite.angle = angle     
        
    def get_moment(self, mass):
        return pymunk.moment_for_box(mass, (self._width, self._height))
    
    def get_shape(self, body):
        return pymunk.Poly.create_box(body, (self._width, self._height))
    
    def _get_top(self):
        return Top_Of_Points(self.get_corners())
    
    def _set_top(self, value: float):
        distance = value - self.top
        
        self.sprite.center_y += distance
        
    top = property(_get_top, _set_top)
    
    def _get_bottom(self):
        return Bottom_Of_Points(self.get_corners())
    
    def _set_bottom(self, value: float):
        distance = value - self.bottom
        
        self.sprite.center_y += distance
        
    bottom = property(_get_bottom, _set_bottom)
    
    def _get_left(self):
        return Left_Of_Points(self.get_corners())
    
    def _set_left(self, value: float):
        distance = value - self.left
        
        self.sprite.center_x += distance
        
    left = property(_get_left, _set_left)
    
    def _get_right(self):
        return Right_Of_Points(self.get_corners())
    
    def _set_right(self, value: float):
        distance = value - self.right
        
        self.sprite.center_x += distance
        
    right = property(_get_right, _set_right)
    
    def _get_width(self):
        return self._width
    
    def _set_width(self, value: float):
        if self._width is not value:
            self._width = value
            
            self.update()
            
            if self.pymunk.engine is not None:
                self.pymunk.engine.refresh(self)            
    
    width = property(_get_width, _set_width)
    
    def _get_height(self):
        return self._height
    
    def _set_height(self, value: float):
        if self._height is not value:
            self._height = value
            
            self.update()
            
            if self.pymunk.engine is not None:
                self.pymunk.engine.refresh(self)            
    
    height = property(_get_height, _set_height)
    
    def _get_color(self):
        if self._color1 == self._color2 == self.color3 == self.color4:
            return self._color1
        
        return self._color1, self._color2, self._color3, self._color4
    
    def _set_color(self, value: arcade.Color):
        if (not self._color1 == self._color2 == self._color3 == self._color4 ==
            value):
            
            self._color1 = value
            self._color2 = value
            self._color3 = value
            self._color4 = value            
        
            self.update()
        
    color = property(_get_color, _set_color)
    
    def _get_color1(self):
        return self._color1
    
    def _set_color1(self, value: arcade.Color):
        if self._color1 is not value:
            self._color1 = value
        
            self.update()
        
    color1 = property(_get_color1, _set_color1)
    
    def _get_color2(self):
        return self._color2
    
    def _set_color2(self, value: arcade.Color):
        if self._color2 is not value:
            self._color2 = value
        
            self.update()
            
    color2 = property(_get_color2, _set_color2)
            
    def _get_color3(self):
        return self._color3
    
    def _set_color3(self, value: arcade.Color):
        if self._color3 is not value:
            self._color3 = value
        
            self.update()
            
    color3 = property(_get_color3, _set_color3)
            
    def _get_color4(self):
        return self._color4
    
    def _set_color4(self, value: arcade.Color):
        if self._color4 is not value:
            self._color4 = value
        
            self.update()
            
    color4 = property(_get_color4, _set_color4)
    
    def _get_outline(self):
        return self._outline
    
    def _set_outline(self, value: float):
        if self._outline is not value:
            self._outline = value
            
            self.update()
    
    outline = property(_get_outline, _set_outline)
                
class Image(Base):
    
    def __init__(self, name: str = arcade.resources.image_key_blue,
                 x: float = 100, y: float = 100,
                 flip_horizontally: bool = False,
                 flip_vertically: bool = False,
                 flip_diagonally: bool = False,
                 rotation: float = 0, collision: str = 'simple',
                 detail: float = 4.5, alpha: float = 255):
        
        if collision == 'box':
            collide = 'None'
            
        else:
            collide = collision.capitalize()
            
        self.sprite = arcade.Sprite(name, center_x = x, center_y = y,
                                    flipped_horizontally = flip_horizontally,
                                    flipped_vertically = flip_vertically,
                                    flipped_diagonally = flip_diagonally,
                                    hit_box_algorithm = collide,
                                    hit_box_detail = detail)

        self.sprite.angle = rotation
                        
        self.name = name
        self.alpha = alpha        
        self.pymunk = Pymunk()
        self.array = None
        
    def get_moment(self, mass):
        return pymunk.moment_for_poly(mass, self.sprite.get_hit_box())
    
    def get_shape(self, body):
        return pymunk.Poly(body, self.sprite.get_hit_box())
        
    def get_corners(self):
        return self.sprite.get_adjusted_hit_box()

    def boundary(self):
        return {'points': self.get_corners(),
                'type': 'polygon'}
    
    def draw(self):
        self.sprite.draw()
    
    def _get_top(self):
        return self.sprite.top
    
    def _set_top(self, value: float):
        self.sprite.top = value
            
    top = property(_get_top, _set_top)
    
    def _get_bottom(self):
        return self.sprite.bottom
    
    def _set_bottom(self, value: float):
        self.sprite.bottom = value
            
    bottom = property(_get_bottom, _set_bottom)
    
    def _get_left(self):
        return self.sprite.left
    
    def _set_left(self, value: float):
        self.sprite.left = value
            
    left = property(_get_left, _set_left)
    
    def _get_right(self):
        return self.sprite.right
    
    def _set_right(self, value: float):
        self.sprite.right = value
            
    right = property(_get_right, _set_right)
    
    def _get_width(self):
        return self.sprite.width
    
    def _set_width(self, value: float):
        self.sprite.width = value
        
        if self.pymunk.engine is not None:
            self.pymunk.engine.refresh(self)        
            
    width = property(_get_width, _set_width)
    
    def _get_height(self):
        return self.sprite.height
    
    def _set_height(self, value: float):
        self.sprite.height = value
        
        if self.pymunk.engine is not None:
            self.pymunk.engine.refresh(self)        
            
    height = property(_get_height, _set_height)
    
    def _get_alpha(self):
        return self.sprite.alpha
    
    def _set_alpha(self, value: float):
        self.sprite.alpha = value
            
    alpha = property(_get_alpha, _set_alpha)
    
    def _get_color(self):
        return self.sprite.color
    
    def _set_color(self, value: float):
        self.sprite.color = value
            
    color = property(_get_color, _set_color)
    
    def _get_collision(self):
        if self.sprite._hit_box_algorithm == 'None':
            return 'box'
            
        return self.sprite._hit_box_algorithm.lower()
    
    def _set_collision(self, value: str):
        if self.collision is not value:
            self.sprite._point_list_cache = None
            self.sprite.points = None
            self.sprite.texture._hit_box_points = None        
        
            if value == 'box':
                self.sprite._texture._hit_box_algorithm = 'None'
            
            else:
                self.sprite._texture._hit_box_algorithm = value.capitalize()
            
    collision = property(_get_collision, _set_collision)
    
    def _get_detail(self):
        return self.sprite.texture._hit_box_detail
    
    def _set_detail(self, value: float):
        if self.detail is not value:
            self.sprite._point_list_cache = None
            self.sprite.points = None
            self.sprite.texture._hit_box_points = None        
            self.sprite.texture._hit_box_detail = value
            
    detail = property(_get_detail, _set_detail)
    
text_id = 0
                
class Text(Base):
    
    def __init__(self, content: str = 'hello', x: float = 100, y: float = 100,
                 color: arcade.Color = (0, 0, 0), size: float = 40,
                 rotation: float = 0):

        self._content = content
        self._color = color
        self._size = size
        self.pymunk = Pymunk()
        self.array = None

        self.update(x, y, rotation)
        
    def draw(self):
        self.sprite.draw()
        
    def boundary(self):
        return {'points': self.get_corners(),
                'type': 'polygon'}
    
    def get_corners(self):
        return self.sprite.get_adjusted_hit_box()
    
    def update(self, x = None, y = None, angle = None):
        global text_id
        
        if x is None: x = self.sprite.center_x
        if y is None: y = self.sprite.center_y
        if angle is None: angle = self.sprite.angle
                
        self.sprite = arcade.draw_text(self._content, x, y, self._color,
                                       self._size, rotation = angle,
                                       align = 'center', anchor_x = 'center',
                                       anchor_y = 'center', bold = text_id)
        text_id += 1
        
    def get_moment(self, mass):
        return pymunk.moment_for_poly(mass, self.sprite.get_hit_box())
    
    def get_shape(self, body):
        return pymunk.Poly(body, self.sprite.get_hit_box())    
        
    def clear(self):
        self.content = ''
    
    def _get_content(self):
        return self._content

    def _set_content(self, text: str):
        if self._content is not text:
            self._content = text
        
            self.update()
            
            if self.pymunk.engine is not None:
                self.pymunk.engine.refresh(self)

    content = property(_get_content, _set_content)
    
    def _get_color(self):
        return self._color
    
    def _set_color(self, value: arcade.Color):
        if self._color is not value:
            self._color = value

            self.update()

    color = property(_get_color, _set_color)
    
    def _get_size(self):
        return self._size
    
    def _set_size(self, value: float):
        if self._size is not value:
            self._size = value

            self.update()
        
            if self.pymunk.engine is not None:
                self.pymunk.engine.refresh(self)

    size = property(_get_size, _set_size)
        
    def _get_width(self):
        return self.sprite.width
    
    def _set_width(self, value: float):
        self.sprite.width = value
        
        if self.pymunk.engine is not None:
            self.pymunk.engine.refresh(self)        
        
    width = property(_get_width, _set_width)
    
    def _get_height(self):
        return self.sprite.height
    
    def _set_height(self, value: float):
        self.sprite.height = value
        
        if self.pymunk.engine is not None:
            self.pymunk.engine.refresh(self)        
        
    height = property(_get_height, _set_height)    

    def _get_top(self):
        return self.sprite.top
    
    def _set_top(self, value: float):
        self.sprite.top = value
        
    top = property(_get_top, _set_top)
    
    def _get_bottom(self):
        return self.sprite.bottom
    
    def _set_bottom(self, value: float):
        self.sprite.bottom = value
        
    bottom = property(_get_bottom, _set_bottom)
    
    def _get_left(self):
        return self.sprite.left
    
    def _set_left(self, value: float):
        self.sprite.left = value
        
    left = property(_get_left, _set_left)
    
    def _get_right(self):
        return self.sprite.right
    
    def _set_right(self, value: float):
        self.sprite.right = value
        
    right = property(_get_right, _set_right)


import math, arcade

from .math import Angle
from .math import Distance
from .math import Direction

def Scale_X_Points(points: arcade.PointList, x_scale: float):
    
    for point in points:
        point[0] *= x_scale
        
    return points
        
def Scale_Y_Points(points: arcade.PointList, y_scale: float):
        
    for point in points:
        point[1] *= y_scale
        
    return points

def Scale_Points(points: arcade.PointList, scale: float):
        
    for point in points:
        point[0] *= scale
        point[1] *= scale
        
    return points    
            
def Rotate_Points(points: arcade.PointList, rotation: float, center_x: float,
                  center_y: float):
    
    points = [list(point) for point in points]
    
    for point in points:
        pos = arcade.rotate_point(point[0], point[1], center_x, center_y,
                                  rotation)
        point[0] = pos[0]
        point[1] = pos[1]
                
    return points
        
def Center_X_Of_Points(points: arcade.PointList):
    center_x = 0
    
    for point in points:
        center_x += point[0]
        
    center_x /= len(points)
    
    return center_x

def Center_Y_Of_Points(points: arcade.PointList):
    center_y = 0
    
    for point in points:
        center_y += point[1]
        
    center_y /= len(points)
    
    return center_y
            
def Left_Of_Points(points: arcade.PointList):
    left = 0
    
    for point in points:
        
        if not points.index(point) == 0:
            
            if point[0] < left:
                left = point[0]
                
        else:
            left = point[0]
            
    return left

def Top_Of_Points(points: arcade.PointList):
    top = 0
    
    for point in points:
        
        if not points.index(point) == 0:
            
            if point[1] > top:
                top = point[1]
                
        else:
            top = point[1]
            
    return top

def Right_Of_Points(points: arcade.PointList):
    right = 0  
    
    for point in points:
        
        if not points.index(point) == 0:
            
            if point[0] > right:
                right = point[0]
                
        else:
            right = point[0]
            
    return right

def Bottom_Of_Points(points: arcade.PointList):
    bottom = 0    
    
    for point in points:
        
        if not points.index(point) == 0:
            
            if point[1] < bottom:
                bottom = point[1]
                
        else:
            bottom = point[1]
            
    return bottom


import arcade, pymunk, math

from .setup import Screen

from .draw import Point
from .draw import Line
from .draw import Strip
from .draw import Shape
from .draw import Circle
from .draw import Box
from .draw import Image
from .draw import Text
from .draw import Positions
from .draw import Position
from .draw import Pymunk

from .math import Distance
from .math import Direction
from .math import Angle

DYNAMIC = pymunk.Body.DYNAMIC
STATIC = pymunk.Body.STATIC
KINEMATIC = pymunk.Body.KINEMATIC
INFINITE = pymunk.inf

sprites = [Point, Line, Strip, Shape, Circle, Box, Image, Text]

class Engine:
    
    def __init__(self, x_gravity: float = 0, y_gravity: float = -1000,
                 damping: float = 1):
        
        Screen.engines.append(self)
        self.space = pymunk.Space()
        self.space.gravity = (x_gravity, y_gravity)
        self.space.damping = damping
        self.space.collision_slop = 1
        self.sprites = []
        self.joints = []
        
    def add(self, sprite, mass: float = 1, body: int = DYNAMIC,
            elasticity: float = None, moment = None, friction: float = 0.2,
            max_x_speed: float = None, max_y_speed: float = None,
            x_gravity: float = None, y_gravity: float = None,
            damping: float = None):
        
        group = type(sprite)
        
        def velocity(body, gravity, damping, dt):
            if sprite.pymunk.damping is not None:
                damping = ((sprite.pymunk.damping * 100.0) / 100.0) ** dt
            
            if sprite.x_gravity is not None:
                gravity = (sprite.x_gravity, gravity[1])
                
            if sprite.y_gravity is not None:
                gravity = (gravity[0], sprite.y_gravity)
                
            pymunk.Body.update_velocity(body, gravity, damping, dt)
                
            if sprite.max_x_speed is not None:
                speed = body.velocity.x
                                
                if abs(speed) > sprite.max_x_speed:
                    speed = sprite.max_x_speed * math.copysign(1, speed)
                    body.velocity = pymunk.Vec2d(speed, body.velocity.y)

            if sprite.max_y_speed is not None:
                speed = body.velocity.y
                                
                if abs(speed) > sprite.max_y_speed:
                    speed = sprite.max_y_speed * math.copysign(1, speed)
                    body.velocity = pymunk.Vec2d(body.velocity.x, speed)        
        
        if group in sprites or group in forces:
            
            sprite.pymunk.moment = moment
            
            if moment is None:
                moment = sprite.get_moment(mass)

            pymunk_body = pymunk.Body(mass, moment, body)
            pymunk_body.position = pymunk.Vec2d(sprite.x, sprite.y)
            pymunk_body.angle = math.radians(sprite.rotation)

            if body == DYNAMIC:
                pymunk_body.velocity_func = velocity
                
            pymunk_shape = sprite.get_shape(pymunk_body)
                        
            if elasticity is not None:
                pymunk_shape.elasticity = elasticity
            
            pymunk_shape.friction = friction
            
            self.space.add(pymunk_body, pymunk_shape)
            self.sprites.append(sprite)
        
            sprite.pymunk.shape = pymunk_shape
            sprite.pymunk.body = pymunk_body
            sprite.pymunk.engine = self
            sprite.pymunk.max_x_speed = max_x_speed
            sprite.pymunk.max_y_speed = max_y_speed
            sprite.pymunk.x_gravity = x_gravity
            sprite.pymunk.y_gravity = y_gravity
            sprite.pymunk.damping = damping            
            
        elif group in joints:
            self.space.add(sprite.joint)
            self.joints.append(sprite) 
        
    def remove(self, sprite):
        if type(sprite) in sprites:
            self.space.remove(sprite.pymunk.body, sprite.pymunk.shape)
            self.sprites.remove(sprite)
            
            sprite.pymunk.engine = None
            sprite.pymunk.shape = None
            sprite.pymunk.body = None
            
            if sprite.pymunk.joint is not None:
                sprite.pymunk.joint.kill()
                
        elif type(sprite) in joints:
            self.space.remove(sprite.joint)
            self.joints.remove(sprite)
            
            self.joint.a.pymunk.joint = None
            self.joint.b.pymunk.joint = None            
        
    def refresh(self, sprite):
        
        mass = sprite.pymunk.body.mass
        moment = sprite.pymunk.moment
        body = sprite.pymunk.body.body_type
        
        if moment is None:
            moment = sprite.get_moment(mass)

        pymunk_body = pymunk.Body(mass, moment, body)
        pymunk_body.position = pymunk.Vec2d(sprite.x, sprite.y)
        pymunk_body.angle = sprite.rotation
        
        if body == DYNAMIC:
            pymunk_body.velocity_func = sprite.pymunk.body.velocity_func
        
        pymunk_shape = sprite.get_shape(pymunk_body)
        pymunk_shape.elasticity = sprite.pymunk.shape.elasticity
        pymunk_shape.friction = sprite.pymunk.shape.friction
        
        self.space.remove(sprite.pymunk.body, sprite.pymunk.shape)
        self.space.add(pymunk_body, pymunk_shape)
    
        sprite.pymunk.shape = pymunk_shape
        sprite.pymunk.body = pymunk_body

    def update(self):
        for sprite in self.sprites:
            sprite.sprite.center_x = sprite.pymunk.shape.body.position.x
            sprite.sprite.center_y = sprite.pymunk.shape.body.position.y
            sprite.sprite.angle = math.degrees(sprite.pymunk.shape.body.angle)
            
            if type(sprite) in forces:
                sprite.calculate(self)
                
        for joint in self.joints:
            if type(joint) == Groove_Joint:
                x1, y1 = Direction(joint.a, Angle(joint.joint.a.position.x,
                                                  joint.joint.a.position.y,
                                                  joint.joint.b.position.x,
                                                  joint.joint.b.position.y))
                
                x2, y2 = Direction(joint.b, Angle(joint.joint.a.position.x,
                                                  joint.joint.a.position.y,
                                                  joint.joint.b.position.x,
                                                  joint.joint.b.position.y))
                x1 += joint.joint.a.position.x
                y1 += joint.joint.a.position.y
                x2 += joint.joint.a.position.x
                y2 += joint.joint.a.position.y
                
                joint.points = ((x1, y1), (x2, y2))
                                
            else:
                joint.points = ((joint.joint.a.position.x,
                                 joint.joint.a.position.y),
                                (joint.joint.b.position.x,
                                 joint.joint.b.position.y))
                        
        self.space.step(1 / Screen.rate)
        
    def collide(self, type1, type2):
        query = self.space.shape_query(type1.pymunk.shape)
        
        for item in query:
            if type2.pymunk.shape == item.shape:
                return True
            
        return False
    
    def _get_x_gravity(self):
        return self.space.gravity.x
    
    def _set_x_gravity(self, value: float):
        self.space.gravity = value, self.space.gravity.y
        
    x_gravity = property(_get_x_gravity, _set_x_gravity)
    
    def _get_y_gravity(self):
        return self.space.gravity.y
    
    def _set_y_gravity(self, value: float):
        self.space.gravity = self.space.gravity.x, value
        
    y_gravity = property(_get_y_gravity, _set_y_gravity)
    
    def _get_damping(self):
        return self.space.damping
    
    def _set_damping(self, value: float):
        self.space.damping = value
        
    damping = property(_get_damping, _set_damping)
    
class Joint(Line):
    
    def __init__(self, sprite1, sprite2, thickness, color):
        
        #if sprite1.pymunk.engine is None or sprite2.pymunk.engine is None:
            #raise Exception('Both sprites should be added to an Engine.')
        
        #if sprite1.pymunk.engine is not sprite2.pymunk.engine:
            #raise Exception('Both sprites should have the same Engine.')
        
        sprite1.pymunk.joint = self
        sprite2.pymunk.joint = self
        
        self._x1 = sprite1.x
        self._y1 = sprite1.y
        self._x2 = sprite2.x
        self._y2 = sprite2.y
        self._thickness = thickness
        self._color1 = color
        self._color2 = color
        self.pymunk = Pymunk()
        self.array = None
        
        self.update(0, 0, 0)    
    
class Pin_Joint(Joint):
    
    def __init__(self, sprite1, sprite2, thickness: float = 1,
                 color: arcade.Color = (0, 0, 0)):
        
        super().__init__(sprite1, sprite2, thickness, color)
        
        self.joint = pymunk.PinJoint(sprite1.pymunk.shape.body,
                                     sprite2.pymunk.shape.body)
        
class Spring_Joint(Joint):
    
    def __init__(self, sprite1, sprite2, length: float = 100,
                 stiffness: float = 300, damping: float = 30,
                 thickness: float = 1, color: arcade.Color = (0, 0, 0)):
        
        super().__init__(sprite1, sprite2, thickness, color)
        
        self.joint = pymunk.DampedSpring(sprite1.pymunk.shape.body,
                                         sprite2.pymunk.shape.body, (0, 0),
                                         (0, 0), length, stiffness, damping)
        
    def _get_length(self):
        return self.joint.rest_length
    
    def _set_length(self, value: float):
        self.joint.rest_length = value
        
    length = property(_get_length, _set_length)
    
    def _get_stiffness(self):
        return self.joint.stiffness
    
    def _set_stiffness(self, value: float):
        self.joint.stiffness = value
        
    stiffness = property(_get_stiffness, _set_stiffness)
    
    def _get_damping(self):
        return self.joint.damping
    
    def _set_damping(self, value: float):
        self.joint.damping = value
        
    damping = property(_get_damping, _set_damping)
        
class Gear_Joint(Joint):
    
    def __init__(self, sprite1, sprite2, angle: float = 1, ratio: float = 1,
                 thickness: float = 1, color: arcade.Color = (0, 0, 0)):
        
        super().__init__(sprite1, sprite2, thickness, color)
        
        self.joint = pymunk.GearJoint(sprite1.pymunk.shape.body,
                                      sprite2.pymunk.shape.body, angle, ratio)
        
    def _get_angle(self):
        return self.joint.phase
    
    def _set_angle(self, value: float):
        self.joint.phase = value
        
    angle = property(_get_angle, _set_angle)
    
    def _get_ratio(self):
        return self.joint.ratio
    
    def _set_ratio(self, value: float):
        self.joint.ratio = value
        
    ratio = property(_get_ratio, _set_ratio)
        
class Groove_Joint(Joint):
    
    def __init__(self, sprite1, sprite2, min_x: float = 100, min_y: float = 100,
                 max_x: float = 200, max_y: float = 200, thickness: float = 1,
                 color: arcade.Color = (0, 0, 0)):
        
        super().__init__(sprite1, sprite2, thickness, color)
        
        self.joint = pymunk.GrooveJoint(sprite1.pymunk.shape.body,
                                        sprite2.pymunk.shape.body,
                                        (min_x, min_y), (max_x, max_y), (0, 0))
        self.a = Distance(0, 0, min_x, min_y)
        self.b = Distance(0, 0, max_x, max_y)
        
    def _get_min_x(self):
        return self.joint.groove_a.x
    
    def _set_min_x(self, value: float):
        self.joint.groove_a.x = value
        
    min_x = property(_get_min_x, _set_min_x)
    
    def _get_min_y(self):
        return self.joint.groove_a.y
    
    def _set_min_y(self, value: float):
        self.joint.groove_a.y = value
        
    min_y = property(_get_min_y, _set_min_y)
    
    def _get_max_x(self):
        return self.joint.groove_b.x
    
    def _set_max_x(self, value: float):
        self.joint.groove_b.x = value
        
    max_x = property(_get_max_x, _set_max_x)
    
    def _get_max_y(self):
        return self.joint.groove_b.y
    
    def _set_max_y(self, value: float):
        self.joint.groove_b.y = value
        
    max_y = property(_get_max_y, _set_max_y)
        
class Pivot_Joint(Joint):
    
    def __init__(self, sprite1, sprite2, pivot_x: float = 100,
                 pivot_y: float = 100, thickness: float = 1,
                 color: arcade.Color = (0, 0, 0)):
        
        super().__init__(sprite1, sprite2, thickness, color)
        
        self.joint = pymunk.PivotJoint(sprite1.pymunk.shape.body,
                                       sprite2.pymunk.shape.body,
                                       (pivot_x, pivot_y))
        
    def _get_pivot_x(self):
        return self.joint.anchor_a.x
    
    def _set_pivot_x(self, value: float):
        self.joint.anchor_a.x = value
        
    pivot_x = property(_get_pivot_x, _set_pivot_x)
    
    def _get_pivot_y(self):
        return self.joint.anchor_a.y
    
    def _set_pivot_y(self, value: float):
        self.joint.anchor_a.y = value
        
    pivot_y = property(_get_pivot_y, _set_pivot_y)    
        
class Motor_Joint(Joint):
    
    def __init__(self, sprite1, sprite2, speed: float = 5,
                 thickness: float = 1, color: arcade.Color = (0, 0, 0)):
        
        super().__init__(sprite1, sprite2, thickness, color)
        
        self.joint = pymunk.SimpleMotor(sprite1.pymunk.shape.body,
                                        sprite2.pymunk.shape.body, speed)
        
    def _get_speed(self):
        return self.joint.rate
    
    def _set_speed(self, value: float):
        self.joint.rate = value
        
    speed = property(_get_speed, _set_speed)
            
joints = [Pin_Joint, Spring_Joint, Gear_Joint, Groove_Joint, Pivot_Joint,
          Motor_Joint]

class Gravity_Force(Circle):
    
    def __init__(self, x: float = 100, y: float = 100, strength: float = 10000,
                 pull: float = 1, size: float = 50,
                 color: arcade.Color = (0, 0, 0), outline: float = 0):
        
        super().__init__(x, y, size, color, outline)
        
        self._strength = strength
        self._pull = pull
                
    def _get_strength(self):
        return self._strength
    
    def _set_strength(self, value: float):
        self._strength = value
        
    strength = property(_get_strength, _set_strength)
    
    def _get_pull(self):
        return self._pull
    
    def _set_pull(self, value: float):
        self._pull = value
        
    pull = property(_get_pull, _set_pull)    
        
    def calculate(self, engine):
        for sprite in engine.sprites:
                        
            if sprite is not self:
                x, y = Direction(
                    self._strength / (sprite.distance(self) / self._pull),
                    sprite.angle(self))
                
                sprite.x_speed += x
                sprite.y_speed += y
        
forces = [Gravity_Force]


import arcade, math, random

def Distance(x1: float, y1: float, x2: float, y2: float):
    return arcade.get_distance(x1, y1, x2, y2)

def Angle(x1: float, y1: float, x2: float, y2: float):
    x, y = x2 - x1, y2 - y1
    radians = math.atan2(y, x)
    radians %= 2 * math.pi
    degrees = math.degrees(radians)
    
    return degrees

def Direction(length: float, angle: float):
    x = length * math.cos(math.radians(angle))
    y = length * math.sin(math.radians(angle))
    
    return x, y

def Random(int1: int, int2: int):
    if int1 > int2:
        return random.randint(int2, int1)

    return random.randint(int1, int2)
    
def Random_Angle():
    return Random(1, 360)

def Random_Color():
    return (Random(0, 255), Random(0, 255), Random(0, 255))

def Random_Alpha_Color():
    return (Random(0, 255), Random(0, 255), Random(0, 255), Random(0, 255))
    
def Pause(time: float):
    arcade.pause(time)


from JoBase import *

# JoBase Car Physics Example

WHEEL_FRICTION = 0.9
WHEEL_SIZE = 30

CAR_X = 600
CAR_Y = 300
CAR_LENGTH = 100
CAR_HEIGHT = 30
CAR_SPEED = 20

SUSPENTION_MIN = 30
SUSPENTION_MAX = 60
SUSPENTION_ANGLE = 50
SUSPENTION_STIFFNESS = 200
SUSPENTION_DAMPING = 5

# Change the values above to customize your car.

class WHEEL:
    # This class handles the wheels of the car.
    
    def __init__(self, body, pos):
        min_x, min_y = Direction(-SUSPENTION_MIN, SUSPENTION_ANGLE)
        max_x, max_y = Direction(-SUSPENTION_MAX, SUSPENTION_ANGLE)
        
        self.wheel = Circle(body.x + max_x * pos, body.y + max_y, WHEEL_SIZE)
        engine.add(self.wheel, friction = WHEEL_FRICTION)
        
        # We add a box so that you can see the wheels rotating.
        self.box = Box(self.wheel.x, self.wheel.y, 30, 5, ANDROID_GREEN)
        
        # The groove joint keeps the wheels in place.
        self.groove = Groove_Joint(body, self.wheel, min_x * pos, min_y, max_x * pos, max_y, color = RED)
        
        # The spring joint gives the car suspention.
        self.spring = Spring_Joint(body, self.wheel, SUSPENTION_MAX, SUSPENTION_STIFFNESS, SUSPENTION_DAMPING)
        
        # The motor joint makes the wheels turn.
        self.motor = Motor_Joint(body, self.wheel, 0)
        
        # We add everything to the physics engine.
        engine.add(self.groove)
        engine.add(self.spring)
        engine.add(self.motor)
        
    def update(self):
        # This function updates the rotation of the green box on the wheel.
        
        self.box.x = self.wheel.x
        self.box.y = self.wheel.y
        self.box.rotation = self.wheel.rotation
        
    def draw(self):
        self.wheel.draw()
        self.box.draw()
        self.groove.draw()

class CAR:
    
    def __init__(self):
        self.body = Box(CAR_X, CAR_Y, CAR_LENGTH, CAR_HEIGHT, RED)
        engine.add(self.body)
        
        # The code below makes the front and rear wheels.
        self.front = WHEEL(self.body, 1)
        self.rear = WHEEL(self.body, -1)
        
    def draw(self):
        self.body.draw()
        self.front.draw()
        self.rear.draw()
        
    def update(self):
        self.front.update()
        self.rear.update()
        
    def set_speed(self, speed):
        # Change the speed of the car.
        self.front.motor.speed = speed
        self.rear.motor.speed = speed

engine = Engine()
car = CAR()

# Make the black bump shape.
bump = Shape(((0, 0), (100, 50), (400, 60), (500, 0)))
bump.y = 200

# We make the bump really heavy so that the car can't push it.
engine.add(bump, mass = 10, friction = WHEEL_FRICTION)

ground = Box(400, 50, 1000, 20)
engine.add(ground, body = STATIC, friction = WHEEL_FRICTION)

def loop():
    
    car.update()
    car.draw()
        
    ground.draw()
    bump.draw()
                
    if Key.left: car.set_speed(-CAR_SPEED)
    elif Key.right: car.set_speed(CAR_SPEED)
    else: car.set_speed(0)
    
    if Screen.resize:
        ground.x = Screen.width / 2
        ground.width = Screen.width

run()


from JoBase import *

# JoBase Joint Physics Example

class BORDERS(List):
    # This class handles all the borders on the screen.
    
    def __init__(self):
        super().__init__()
        
        coordinates = self.get_coordinates()
        
        # First, we make all the borders and add them to the list.
        
        for i in range(int(len(coordinates) / 4)):
            line = Line(coordinates[i * 4], coordinates[i * 4 + 1], coordinates[i * 4 + 2], coordinates[i * 4 + 3], 10, GRAY)
            self.append(line)
            engine.add(line, body = STATIC, elasticity = 0.9)
            
    def resize(self):
        # When the screen is resized, this function is called.
        # This function re-positions all the line borders.
        
        coordinates = self.get_coordinates()
        
        for i, line in enumerate(self):
            line.points = ((coordinates[i * 4], coordinates[i * 4 + 1]), (coordinates[i * 4 + 2], coordinates[i * 4 + 3]))
            
    def get_coordinates(self):
        # This returns all the coordinates for each line border.
        
        return [0, Screen.height / 2, Screen.width, Screen.height / 2,
                Screen.width / 3, Screen.height, Screen.width / 3, 0,
                Screen.width / 3 * 2, Screen.height, Screen.width / 3 * 2, 0,
                0, 0, 0, Screen.height,
                Screen.width, 0, Screen.width, Screen.height,
                0, Screen.height, Screen.width, Screen.height,
                0, 0, Screen.width, 0]
    
class OBJECTS():
    # This class handles all the boxes and joints on the screen.
    
    def __init__(self):        
        self.joints = List()
        self.boxes = List()
        
        # Set up the names of all the joints and their values.
        
        names = ['Pin_Joint(box_1, box_2)',
                 'Spring_Joint(box_1, box_2, damping = 1)',
                 'Gear_Joint(box_1, box_2)',
                 'Groove_Joint(box_1, box_2, 50, 50, 150, 150)',
                 'Pivot_Joint(box_1, box_2, box_2.x + 30, box_2.y + 30)',
                 'Motor_Joint(box_1, box_2, 20)']
        
        x = Screen.width / 6
        y = Screen.height / 4 * 3
        
        for name in names:
            # First we make the boxes.
            
            box_1 = Box(x - 50, y, 50, 30, Random_Color(), 0, Random_Angle())
            box_2 = Box(x + 50, y, 50, 30, Random_Color(), 0, Random_Angle())
            
            if name.startswith('Pivot'): body = STATIC
            else: body = DYNAMIC
                
            # Add the boxes to the engine and the list.
            engine.add(box_1, elasticity = 0.9, body = body)
            engine.add(box_2, elasticity = 0.9)
            self.boxes.append(box_1)
            self.boxes.append(box_2)
            
            # Make the joint between the boxes.
            joint = eval(name)
            joint.thickness = 5
            joint.color1 = box_1.color
            joint.color2 = box_2.color
            engine.add(joint, elasticity = 0.9)
            self.joints.append(joint)
            
            # Change the positions for the next set of boxes.
            x += Screen.width / 3
            
            if x > Screen.width / 6 * 5:
                x = Screen.width / 6
                y = Screen.height / 4
            
    def draw(self):
        self.joints.draw()
        self.boxes.draw()
        
class TEXT(List):
    # This class handles the text on the screen.
    
    def __init__(self):
        super().__init__()
        
        # Set up the names of all the joints.
        names = ['Pin', 'Spring', 'Gear', 'Groove', 'Pivot', 'Motor']
        
        x = Screen.width / 6
        y = Screen.height / 4 * 3
        
        for name in names:
            # Make the text and add it to the list.
            
            text = Text(name + ' Joint', x, y)
            self.append(text)
            
            # Change the positions for the next piece of text.
            x += Screen.width / 3
            
            if x > Screen.width / 6 * 5:
                x = Screen.width / 6
                y = Screen.height / 4
                
    def resize(self):
        # When the screen is resized, this function is called.
        # This function re-positions all the text.        
        
        x = Screen.width / 6
        y = Screen.height / 4 * 3
        
        for text in self:
            text.x = x
            text.y = y
            
            x += Screen.width / 3
            
            if x > Screen.width / 6 * 5:
                x = Screen.width / 6
                y = Screen.height / 4            

engine = Engine()
objects = OBJECTS()
borders = BORDERS()
text = TEXT()

Screen.color = LIGHT_GRAY
Mouse.box = None

def loop():
    
    if Screen.resize: resize()
    
    for box in objects.boxes:
        if Mouse.press and Mouse.collide(box):
            Mouse.box = box
            
    if Mouse.box is not None:
        # Move the box to the mouse position.
        
        Mouse.box.x = Mouse.x + Mouse.x_change
        Mouse.box.y = Mouse.y + Mouse.y_change
        
    if Mouse.release: Mouse.box = None
    
    borders.draw()
    text.draw()
    objects.draw()
    
def resize():
    # Resize the borders and text.
    # We don't bother to re-position the boxes.
    
    borders.resize()
    text.resize()
        
run()


import arcade, os

class List:
    
    def __init__(self, *objects):
        self.array = list(objects)
            
    def __len__(self):
        return len(self.array)
    
    def __iter__(self):
        return iter(self.array)
    
    def __str__(self):
        return str(self.array)
    
    def __repr__(self):
        return repr(self.array)    

    def __getitem__(self, key):
        return self.array[key]
    
    def __setitem__(self, key, value):
        self.array[key] = value
    
    def draw(self):
        for element in self.array:
            if 'draw' in dir(element):
                element.draw()
                
    def append(self, element):
        element.array = self
        self.array.append(element)
        
    def clear(self):
        self.array.clear()        
        
    def remove(self, element):
        element.array = None
        self.array.remove(element)
    
    def add(self, element):
        # It does exactly the same thing as append.
        self.append(element)
        
    def index(self, element):
        return self.array.index(element)        
        
    def collide(self, other):
        return collide(self.boundary(), other.boundary())    
        
    def boundary(self):
        boundaries = []
        
        for item in self.array:
            boundaries.append(item.boundary())
            
        return {'list': boundaries,
                'type': 'list'}
    
class Content:
    
    def __init__(self, array, name):
        self.array = list(array)
        self.name = name
        
    def __getitem__(self, key):
        return self.array[key]    
        
    def __setitem__(self, key, value):
        self.array[key] = value
        
        file = open(self.name, 'r')
        array = file.readlines()
        file.close()
        
        array[key] = str(value) + '\n'
                
        file = open(self.name, 'w')
        file.writelines(array)
        file.close()
    
    def __iter__(self):
        return iter(self.array)
    
    def __str__(self):
        return str(self.array)
    
    def __repr__(self):
        return repr(self.array)
    
    def __len__(self):
        return len(self.array)
    
    def append(value):
        self.array.append(value)
    
class File:
    
    def __init__(self, name: str = 'text_file.txt'):
        # Python automatically creates a new text file if it doesn't exist.

        if not os.path.isfile(name):
            file = open(name, 'w').close()
        
        self._name = name
        self.content = Content(self.read(), name)
        
    def clear(self):
        open(self.name, 'w').close()
        self.content.array = []
    
    def add(self, *items):
        file = open(self.name, 'a')
        
        for item in items:
            file.write(str(item) + '\n')
            self.content.append(item)
            
        file.close()
        
    def write(self, *items):
        file = open(self.name, 'w')
        self.content.array = []
        
        for item in items:
            file.write(str(item) + '\n')
            self.content.append(item)
            
        file.close()
        
    def delete(self):
        os.remove(self.name)
        del self
        
    def read(self):
        file = open(self.name, 'r')
        
        read = file.readlines()
        array = []
    
        for element in read:
            element = element.strip()
        
            if element.replace('.', '', 1).isdigit():
                element = eval(element)
            
            array.append(element)
        
        file.close()
    
        return array
    
    def _get_name(self):
        return self._name
    
    def _set_name(self, name: str):
        self.content.name = name
        os.rename(self._name, name)
        self._name = name
        
    name = property(_get_name, _set_name)

class Sound:
    
    def __init__(self, name: str = 'collect.wav'):
        self.name = name
        self.sound = arcade.load_sound(name)
        self.volume = 1
        self.loop = False
        
    def play(self):
        arcade.play_sound(self.sound, self.volume, 0, self.loop)


import math, warnings, arcade

from arcade import PointList

def collide(type1: dict, type2: dict):
    # The code below checks every combination of collision.
    
    if type1['type'] == 'none' or type2['type'] == 'none':
        warnings.warn("You entered a shape that doesn't have collision.")
        
    elif type1['type'] == 'point' and type2['type'] == 'point':
        
        return rect_to_point_collision(type1['x'], type1['y'], type2['x'],
                                       type2['y'])
    
    elif type1['type'] == 'polygon' and type2['type'] == 'polygon':
        
        return polygon_to_polygon_collision(type1['points'], type2['points'])
    
    elif type1['type'] == 'point' and type2['type'] == 'polygon':
        
        return point_to_polygon_collision(type1['x'], type1['y'],
                                          type2['points'])
    
    elif type1['type'] == 'polygon' and type2['type'] == 'point':
        
        return point_to_polygon_collision(type2['x'], type2['y'],
                                          type1['points'])
    
    elif type1['type'] == 'list' and type2['type'] == 'list':
        
        return list_to_list_collision(type1['list'], type2['list'])
        
    elif (type1['type'] == 'list' and type2['type'] == 'point' or
          type1['type'] == 'list' and type2['type'] == 'polygon'):
        
        return list_to_object_collision(type1['list'], type2)
    
    elif (type1['type'] == 'point' and type2['type'] == 'list' or
          type1['type'] == 'polygon' and type2['type'] == 'list'):
                
        return list_to_object_collision(type2['list'], type1)
    
def point_to_point_collision(x1: float, y1: float, x2: float, y2: float):
    return x1 == x2 and y1 == y2

def polygon_to_polygon_collision(points1: PointList, points2: PointList):
    return arcade.are_polygons_intersecting(points1, points2)

def point_to_polygon_collision(x1: float, y1: float, points2: PointList):
    return arcade.is_point_in_polygon(x1, y1, points2)

def list_to_object_collision(list1: list, boundary2: dict):
    for boundary in list1:
        if collide(boundary, boundary2):
            return True
        
def list_to_list_collision(list1: list, list2: list):
    for boundary1 in list1:
        for boundary2 in list2:
            if collide(boundary1, boundary2):
                return True


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "JoBase",
    version = "1.1",
    author = "Reuben Grey Ford",
    author_email = "info@jobase.org",
    description = "Python Educational Resource",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "http://jobase.org",
    packages = setuptools.find_packages(),
    install_requires = ['arcade'],    
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


"""
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

# Welcome to JoBase.
# This is a Python educational resource for beginner coders.

import arcade, os

from arcade.resources import *
from .color import *

from .setup import Screen
from .setup import Mouse
from .setup import Key

from .engine import DYNAMIC
from .engine import STATIC
from .engine import KINEMATIC
from .engine import INFINITE

from .engine import Engine

from .engine import Pin_Joint
from .engine import Spring_Joint
from .engine import Gear_Joint
from .engine import Groove_Joint
from .engine import Pivot_Joint
from .engine import Motor_Joint

from .engine import Gravity_Force

from .draw import Point
from .draw import Line
from .draw import Strip
from .draw import Shape
from .draw import Circle
from .draw import Box
from .draw import Image
from .draw import Text

from .math import Distance
from .math import Angle
from .math import Direction
from .math import Random
from .math import Random_Angle
from .math import Random_Color
from .math import Random_Alpha_Color

from .data import List
from .data import File
from .data import Sound

from .point import Rotate_Points
from .point import Scale_X_Points
from .point import Scale_Y_Points
from .point import Scale_Points
from .point import Center_X_Of_Points
from .point import Center_Y_Of_Points
from .point import Left_Of_Points
from .point import Top_Of_Points
from .point import Right_Of_Points
from .point import Bottom_Of_Points

print('Welcome to JoBase. Please visit the official website \'jobase.org\' for '
      'tutorials and documentation.')

for name in dir(Screen):
    if name.startswith('CURSOR_'):
        globals()[name[7:]] = getattr(Screen, name)
        
def Random_Screen_X():
    return Random(0, Screen.width)
    
def Random_Screen_Y():
    return Random(0, Screen.height)
    
def Random_Screen_Pos():
    return Random_Screen_X(), Random_Screen_Y()

def run():
    arcade.run()



import arcade, sys

from .collision import collide

from .math import Angle
from .math import Distance
from .math import Direction

module = sys.modules['__main__']

alphabet = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
            'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']

class SCREEN(arcade.Window):
    
    def __init__(self):
        super().__init__(800, 600, 'JoBase', resizable = True,
                         antialiasing = False)
                
        # The code above gets all the colors stored in 'arcade.color' and makes
        # global variables with the same name.        
        
        self.engines = []
        self.module = module
        self.resize = 0
        self.time = 0
        self.quit = 0
        self._color = (255, 255, 255)
        self._rate = 60
        self._fullscreen = 0
        self._visible = 1
        
        arcade.set_background_color(self._color)
        self.set_update_rate(1 / 60)
                
        # The code above finds all the variables starting with 'CURSOR_' and
        # makes new global variables without the 'CURSOR_' prefix.
                
    def collide(self, other):
        # This function checks if 'other' is on the screen.
        return collide(self.boundary(), other.boundary())
        
    def boundary(self):
        # This dictionary is eventually passed into the collision function.
        
        return {'points': arcade.get_rectangle_points(self.width / 2,
                                                      self.height / 2,
                                                      self.width, self.height),
                'type': 'polygon'}
    
    def _get_visible(self):
        return self._visible
    
    def _set_visible(self, value: bool):
        # 'set_visible' is an arcade function.
        self.set_visible(value)
        self._visible = value
        
    visible = property(_get_visible, _set_visible)
    
    def _get_x(self):
        # 'get_location' is an arcade function.
        return self.get_location()[0]
    
    def _set_x(self, value: float):
        # 'set_location' is an arcade function.
        self.set_location(value, self.y)
        
    x = property(_get_x, _set_x)
        
    def _get_y(self):
        return self.get_location()[1]
    
    def _set_y(self, value: float):
        self.set_location(self.x, value)
        
    y = property(_get_y, _set_y)
    
    def _get_title(self):
        return self.caption
    
    def _set_title(self, value: str):
        self.set_caption(value)
        
    title = property(_get_title, _set_title)
    
    def _get_fullscreen(self):
        return self._fullscreen
    
    def _set_fullscreen(self, value: bool):
        self.set_fullscreen(value)
        self.resize = 1
        # 'resize' becomes true when it enters fullscreen mode.
        
    fullscreen = property(_get_fullscreen, _set_fullscreen)
    
    def _get_rate(self):
        return self._rate
    
    def _set_rate(self, value: float):
        # We write '1 / value' to make the user set the frames per second.
        self.set_update_rate(1 / value)
        self._rate = value
        
    rate = property(_get_rate, _set_rate)
    
    def _get_color(self):
        return self._color
    
    def _set_color(self, value: arcade.Color):
        arcade.set_background_color(value)
        self._color = value
        
    color = property(_get_color, _set_color)
        
    def centralize(self):
        # A JoBase user could just use 'center_window'.
        self.center_window()
        
    def exit(self):
        self.quit = 1
        
    def on_update(self, time):
        self.time = 1 / time
        
        arcade.start_render()
        
        if 'loop' in dir(module):
            module.loop()
            
        for engine in self.engines:
            engine.update()
        
        self.resize = 0
        Key.press = 0
        Key.release = 0
        Mouse.press = 0
        Mouse.release = 0
        Mouse.up = 0
        Mouse.down = 0
        Mouse.move = 0
        Mouse.scroll = 0
        
        if self.quit:
            arcade.close_window()
            
        """
        When the user presses the close button, the 'module.loop()' function
        runs once before the window closes.
        """
        
    def on_close(self):
        self.quit = 1
        
    def on_mouse_motion(self, x, y, dx, dy):
        Mouse.move = 1
        Mouse.x_change = x - Mouse._x
        Mouse.y_change = y - Mouse._y
        Mouse._x = x
        Mouse._y = y        

    def on_mouse_scroll(self, x, y, button, direction):
        Mouse.scroll = 1
        
        if direction == 1.0:
            Mouse.scroll_up = 1
            
        elif direction == -1.0:
            Mouse.scroll_down = 1

    def on_mouse_press(self, x, y, button, modifiers):
        Mouse.press = 1
        
        if button == 1:
            Mouse.left = 1
            
        elif button == 4:
            Mouse.right = 1
            
        elif button == 2:
            Mouse.middle = 1

    def on_mouse_release(self, x, y, button, modifiers):
        Mouse.release = 1
        
        if button == 1:
            Mouse.left = 0
            
        elif button == 4:
            Mouse.right = 0
            
        elif button == 2:
            Mouse.middle = 0
        
    def on_resize(self, width, height):
        super().on_resize(width, height)
        self.resize = 1

    def on_key_press(self, key, modifiers):
        Key.press = 1
        
        for keys in dir(arcade.key):
            if not keys.startswith('__') and key == getattr(arcade.key, keys):
                keys = keys.lower()
                setattr(Key, keys, 1)
                
                if keys in alphabet:
                    if Key.lshift or Key.rshift:
                        Key.text += keys.upper()
                        
                    else:
                        Key.text += keys
                        
                elif keys == 'backspace':
                    Key.text = Key.text[:-1]
                    
                elif keys.startswith('num_') or keys.startswith('key_'):
                    Key.text += keys[-1]
                
        # The code above finds all the values in 'arcade.key'. If a key is true,
        # we set the same JoBase key to true.

    def on_key_release(self, key, modifiers):
        Key.release = 1
        
        for keys in dir(arcade.key):
            if not keys.startswith('__') and key == getattr(arcade.key,
                                                            str(keys)):
                keys = keys.lower()
                setattr(Key, str(keys), 0)
                
class KEY:
    
    def __init__(self):
        self.press = 0
        self.release = 0
        self.text = ''
        
        for keys in dir(arcade.key):
            if not keys.startswith('__'):
                keys = keys.lower()
                setattr(self, str(keys), 0)
                
        # The code above finds all the values in 'arcade.key' and makes a
        # JoBase key with the same name (lowercase).
        
    def clear(self):
        self.text = ''
        
class MOUSE():
    
    def __init__(self):
        
        super()
        
        self.press = 0
        self.release = 0
        self.left = 0
        self.right = 0
        self.middle = 0
        self.scroll_up = 0
        self.scroll_down = 0
        self.move = 0
        self.scroll = 0
        self._x = 0
        self._y = 0
        self.x_change = 0
        self.y_change = 0
        self._visible = 1
        self._cursor = None
        
    def boundary(self):
        return {'x': self._x, 'y': self._y, 'type': 'point'}
    
    def _get_visible(self):
        return self._visible
    
    def _set_visible(self, value: bool):
        Screen.set_mouse_visible(value)
        self._visible = value
        
    visible = property(_get_visible, _set_visible)
    
    def _get_cursor(self):
        return self._cursor
    
    def _set_cursor(self, name):
        Screen.set_mouse_cursor(Screen.get_system_mouse_cursor(name))
        self._cursor = name
        
    cursor = property(_get_cursor, _set_cursor)
        
    def _get_x(self):
        return self._x
    
    def _set_x(self, value: float):
        Screen.set_mouse_position(round(value), round(self._y))
        
        self._x = value
        
    x = property(_get_x, _set_x)
    
    def _get_y(self):
        return self._y
    
    def _set_y(self, value: float):
        Screen.set_mouse_position(round(self._x), round(value))
        
        self._y = value
        
    y = property(_get_y, _set_y)
    
    def collide(self, other):
        return collide(self.boundary(), other.boundary())
    
    def angle(self, other):
        return Angle(self.x, self.y, other.x, other.y)

    def distance(self, other):
        return Distance(self.x, self.y, other.x, other.y)    
    
Key = KEY()
Mouse = MOUSE()
Screen = SCREEN()